/* $Id: vga.c,v 1.18 1999/06/27 16:01:15 marcus Exp $
******************************************************************************

   IBM VGA chipset driver

   Copyright (C) 1996-1997	Jason McMullan	[jmcc@ggi-project.org]
   Copyright (C) 1998		Brian S. Julin	[bri@forcade.calyx.com]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

   This code was derived from the following sources of information:
   [1] Xfree86 VGA driver
   [2] Zen of Graphics Programming (Michael Abrash)
   [3] VGADOC4b by Finn Thoegersen

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Jon_Taylor
#define DRIVER_NAME	"IBM VGA chipset driver"
#define DRIVER_REV	"$Revision: 1.18 $"

#include <kgi/module.h>
#include <IBM/vga.h>
#include "vga_io.inc"

#define CURSOR_HW	/* Hardware cursor */
#define	POINTER_FONT	/* Font-hack mouse cursor in textmodes */
#define SLOW_POINTER	/* #define this if your pointer flickers */

struct kgi_vgapriv {
  struct mem_region curr_mmio;
};

static struct kgi_chipset chipinfo =
{
	"IBM",			/* Manufacturer				*/
	"VGA",			/* Model				*/
	GM_TEXT16 | GM_1BIT |	/* Available modes			*/
	GM_4BIT | GM_8BIT,
	0,			/* Flags				*/
	{ 920, 550 },		/* Maximum resolution			*/
	NULL,			/* Private data hook			*/
	0,		        /* RAMsize - filled in during detection */
	{ 25000000, 32000000 }	/* DCLK limits				*/
};

/* This table holds the default values for the VGA registers.  A copy	*/
/* of it will be filled in as appropriate during mode programming in	*/
/* kgim_chipset_set_mode().						*/
static struct statetable default_state =
{
	0x23,							/* MISC	*/
	{
		0x03, 0x00, 0x00, 0x00, 0x00			/* SEQ	*/
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* CRT	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x1x	*/
		0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC	*/
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC	*/
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x20, 0x00, 0x0F, 0x00, 0x00
	}
};

struct statetable reset_state;
kgi_modereq default_kgi_modereq;
struct kgi_mode default_kgi_mode;

static unsigned long fb_text_len;
static unsigned long fb_text;
static unsigned long fb_graph_len;
static unsigned long fb_graph;

static struct io_region vga_io;

static int detect_VGA(struct kgi_chipset *chip)
{
	int err;
	kgiu8 temp1, temp2;

	vga_io.baseptr        =       IO_NULL;
	vga_io.decode         =       VGA_IO_DECODE_MASK;
	vga_io.name           =       "IBM VGA (probing)";
	
	/* Now try to claim VGA ATC/SEQ/GRC PIO area so we can probe it. */
	vga_io.base           =       ATCI;
	vga_io.size           =       0x10;

#define NEED_REGION \
  if ((err = io_check_region(&vga_io))) { \
    DEBUG0("Giving up: PIO[%x-%x] region not available for probe.\n", \
           vga_io.base, (vga_io.base + vga_io.size - 1)); \
    return err; }; \
  io_claim_region(&vga_io);

        NEED_REGION;

	/* Check for read/writeable SEQI -- should reject an EGA.
	   Certainly hope so because the next test will hose it up
	   otherwise.  Anyone got a more EGA-friendly way? */
	temp1 = io_in8(SEQI);
	io_out8(0, SEQI);
	if ((io_in8(SEQI) & VGA_SEQI_MASK) ||
	    (io_out8(SEQ_MEMMODE, SEQI),
	     ((io_in8(SEQI) & VGA_SEQI_MASK) != SEQ_MEMMODE))) {
	        DEBUG0("Giving up: no response from"
		       "sequencer index at PIO[%x].\n", SEQI);
		io_free_region(&vga_io);
		return(-E(CHIPSET, FAILED));
	}
	io_out8(temp1, SEQI);

	/* Check for MISCr, MISCw quirk -- should be a good test for a VGA. */
	temp1 = io_in8(MISCw);
	/* Schroedinger's principle at work... */
	reset_state.MISC = temp2 = io_in8(MISCr);
	temp2 = temp2 ^ (MISC_ENB_RAM | MISC_PAGE_SEL);
	io_out8(temp2, MISCw);
	if ((io_in8(MISCr) != temp2) ||
	    ((io_in8(MISCw) != temp1) && (io_in8(MISCw) == temp2))) {
	        DEBUG0("Giving up: MISC r/w test failed.\n");
		io_free_region(&vga_io);
		return(-E(CHIPSET, FAILED));
	}

	temp2 ^= (MISC_ENB_RAM | MISC_PAGE_SEL); /* put old value back. */
	io_out8(temp2, MISCw);

#define IS_MONO_IO (vga_io.base == VGA_MONO_PIO_BASE)

	/* Now try to grab the rest of PIO -- don't alter the existing
	   configuration unless PIO base is occupied. */
	io_free_region(&vga_io);
	vga_io.size = 0x20;
	if (!(temp2 & MISC_COLOR_IO)) vga_io.base = VGA_MONO_PIO_BASE;
	if ((err = io_check_region(&vga_io))) {
	        vga_io.base = IS_MONO_IO ? ATCI : VGA_MONO_PIO_BASE;
	        DEBUG1("Preset IO region not available,"
		       "switching IO base.");
	}

	NEED_REGION;

	/* Program the adaptor to use the available IO base if necessary. */
	if ((!(temp2 | MISC_COLOR_IO)) != IS_MONO_IO) {
	  /* TODO: This adjusts the IO base for the VGA PIO macros. */
	  /* We should change that behavior so as not to obfuscate it. */
	  /* IO base should be stored in per-chip private data area. */
	        outbMISC((temp2 & ~MISC_COLOR_IO) |
			 (IS_MONO_IO ? 0 : MISC_COLOR_IO));
	        DEBUG1("Set IO base to PIO[%x]\n", vga_io.base);
	}


#ifndef NOT_QUITE_VGA_COMPATIBLE
	/* Check for FCTRL quirk.  This ensures the CRT we have really
	   belongs to us and not to another card. */

	temp1 = io_in8(FCTRLr);
	io_out8(0x0, (IS_MONO_IO ? FCTRLwm : FCTRLwc));
	while ((io_in8(FCTRLr) & 0x8) ||
	       (io_out8(0x8, (IS_MONO_IO ? FCTRLwm : FCTRLwc)),
		!(io_in8(FCTRLr) & 0x8))) {
	        /* Check if we ever had to switch PIO base before. */
	        if (!(temp2 | MISC_COLOR_IO) != IS_MONO_IO) {
	                  DEBUG0("Giving up: FCTRL r/w test failed"
				 "and no alternate PIO.\n");
	                  DEBUG1("Multiple active 6854 based adaptors??\n");
			  outbMISC(temp2);
			  io_free_region(&vga_io);
			  return(-E(CHIPSET, FAILED));
		};
		/* Try to switch PIO base. */
		io_free_region(&vga_io);
		vga_io.base = IS_MONO_IO ? ATCI : VGA_MONO_PIO_BASE;
		NOTICE1("Maybe two 6854 based adaptors."
			"Trying to switch IO base.\n");
		NEED_REGION;
		/* Run the FCTRL test again.
		   Failure this time will be fatal. */
		outbMISC((temp2 & ~MISC_COLOR_IO) |
			 (IS_MONO_IO ? 0 : MISC_COLOR_IO));
		DEBUG1("Switched IO base to PIO[%x]\n", vga_io.base);
		temp1 = io_in8(FCTRLr);
		io_out8(0x0, (IS_MONO_IO ? FCTRLwm : FCTRLwc));
	}
	/* Restore old value */
	io_out8(temp1, (IS_MONO_IO ? FCTRLwm : FCTRLwc));
#endif

	/* I've noticed that setting bit 3 of FCTRL seems
	   to alter the value of MISC bit 2 on some (all?) chipsets.
	   Anyone know anything about this?  Is it a standard feature
	   that could be used to clinch VGA ID?   Will it ever confound 
	   the above test in some cases that outbMISC() is 
	   performed?   Ah, the things I can think of to keep
	   me awake late at night.  -- BSJ */

	io_free_region(&vga_io);
	vga_io.name = IS_MONO_IO ? "IBM VGA mono PIO" : "IBM VGA color PIO";

#undef NEED_REGION
#undef IS_MONO_IO

	/* better memory test would be nice -- actually read/write
	   maybe to detect 128K cards.  Also this assumes
	   SR04_256K_ACCESS has not been altered before insertion */
	chip->ramsize = (inbSEQ(SEQ_MEMMODE) & SR04_256K_ACCESS) ?
		(256 * 1024) : (64 * 1024);

	return EOK;
}

/* GGI interface */

/* Mode specific frame buffers */

#ifdef POINTER_FONT
# include "vga_text16_enh.inc"
#else
# include "vga_text16.inc"
#endif

#include "vga_fb.inc"

#ifdef DEBUG_FB
# include "vga_debug_fb.inc"
#endif


/* Internal Funtion -- kgi_vga_dpy_init

   Initializes the kgi_display structure handed in by the KGI Manager 
   (e.g. fbcon-kgi) including the private data areas. */
static int kgi_vga_dpy_init (struct kgi_display *dpy) {

  /* Create a kgi_chipset and link it into the dpy */
  ASSERT(dpy);
  ASSERT(dpy->card);
  ASSERT(dpy->card->chipset == NULL);

  if(!(dpy->card->chipset = kgi_malloc_small(sizeof(chipinfo)))) {
    DEBUG0("Error Allocating kgi_chipset structure");
    return(-E(CHIPSET,FAILED));
  };

  memcpy(dpy->card->chipset, &chipinfo, sizeof(chipinfo)); 

  if(!(dpy->card->chipset->priv = kgi_malloc_small(sizeof(struct kgi_vgapriv)))) {
    DEBUG0("Error Allocating private data area.");
    kgi_free_small(dpy->card->chipset);
    return(-E(CHIPSET,FAILED));
  };

  return(-EOK);
  
}

/* Internal Funtion -- kgi_vga_dpy_done

   Destroys the chipset area associated with a kgi_display.

*/
static void kgi_vga_dpy_done (struct kgi_display *dpy) {
  ASSERT(dpy);
  ASSERT(dpy->card);
  ASSERT(dpy->card->chipset);
  ASSERT(dpy->card->chipset->priv);

  if(dpy && dpy->card && dpy->card->chipset) {
    if (dpy->card->chipset->priv) 
      kgi_free_small((struct kgi_vgapriv *)(dpy->card->chipset->priv));
    kgi_free_small(dpy->card->chipset);
  }

}



static void chipset_setup_dpy(struct kgi_display *dpy, struct kgi_mode *mode)
{

ASSERT(dpy);
ASSERT(dpy->card);
ASSERT(dpy->card->chipset);
ASSERT(dpy->card->chipset->priv);

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
		dpy->set_font		= text16_set_font;
		dpy->set_origin		= text16_set_origin;
		dpy->set_split		= text16_set_split;
		dpy->set_frame		= text16_set_frame;
		dpy->show_cursor	= text16_show_cursor;
		dpy->hide_cursor	= text16_hide_cursor;
		dpy->undo_cursor	= text16_undo_cursor;
		dpy->show_pointer	= text16_show_pointer;
		dpy->hide_pointer	= text16_hide_pointer;
		dpy->undo_pointer	= text16_undo_pointer;
		dpy->fb.dpy		= dpy;
		dpy->fb.size.x 		= mode->request.virt.x;
		dpy->fb.size.y		= mode->request.virt.y;
		dpy->fb.bpp		= MAP_16_BPP;
#define VGAMEM (((struct kgi_vgapriv *)(dpy->card->chipset->priv))->curr_mmio)
		dpy->fb.pbuf		= VGAMEM.baseptr;
#undef VGAMEM
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= &text16_color_io;
		KGIM_PRIVATE(dpy, driver) = text16_data_realloc(KGIM_PRIVATE(dpy,driver));
		if (KGIM_PRIVATE(dpy, driver) == NULL)
			INTERNAL_ERROR;
		break;
		default:
		dpy->set_font		= NULL;
		dpy->set_origin		= vgafb_set_origin;
		dpy->set_split		= vgafb_set_split;
		dpy->set_frame		= vgafb_set_frame;
		dpy->show_cursor	= NULL;
		dpy->hide_cursor	= NULL;
		dpy->undo_cursor	= NULL;
		dpy->show_pointer	= NULL;
		dpy->hide_pointer	= NULL;
		dpy->undo_pointer	= NULL;
		dpy->fb.dpy		= dpy;
		dpy->fb.size.x		= 0;
		dpy->fb.size.y		= 0;
		dpy->fb.bpp		= 0;
		dpy->fb.pbuf		= NULL;
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= NULL;
		kgim_accel_set_mode(dpy,mode);
	}
	memcpy(&(dpy->mode), mode, sizeof(struct kgi_mode));
}

/* Takes a mode and a bits-per-pixel value */
static void chipset_setup_mode(struct kgi_display *dpy, 
			       struct kgi_mode *mode)
{
	int bpp=0;

ASSERT(dpy);
ASSERT(dpy->card);
ASSERT(dpy->card->chipset);
ASSERT(dpy->card->chipset->priv);

	/* By default, setup for graphics modes */

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
#define VGAMEM (((struct kgi_vgapriv *)(dpy->card->chipset->priv))->curr_mmio)
		mode->fb	= VGAMEM.baseptr;
		mode->fb_size	= VGAMEM.size;
		mode->fsize	= mode->request.virt.x * 
		                  mode->request.virt.y * 2;
		mode->mem	= &text16_color_mem;
		mode->io	= &text16_color_io;
		return; 	/* Done setting up */
		case KGIGT_1BIT:
		bpp = 1;
		break;
		case KGIGT_4BIT:
		bpp = 4;
		break;
		case KGIGT_8BIT:
		bpp = 8;
		break;
		default:
		ERROR("Unsupported mode in chipset_setup_mode()!");
		INTERNAL_ERROR;
		return;
	}
	
	mode->fb	= VGAMEM.baseptr;
	mode->fb_size	= VGAMEM.size;
#undef VGAMEM
	mode->fsize	= (mode->request.virt.x * 
			   mode->request.virt.y * 8) / bpp;
	mode->mem	= NULL;
	mode->io	= NULL;
}

static int chipset_get_reset_mode(struct kgi_display *dpy)
{
	kgi_graphtype gmode = KGIGT_TEXT16;
	int magx = 1, magy = 1;
	kgiu8 temp;
	int DCLKdiv = 8;
	int dotclock;
	struct kgi_timing tx,ty;

ASSERT(dpy);
ASSERT(dpy->card);
ASSERT(dpy->card->chipset);
ASSERT(dpy->card->chipset->priv);

	temp = reset_state.MISC; /* Stored and altered by probe. */
	save_state(&reset_state);
	reset_state.MISC = temp;

	if (reset_state.ATC[0x10] & AR10_GRAPHICS) {
		switch (reset_state.GRC[0x05] & GR05_WRITEMODE3) {
			case GR05_WRITEMODE0:
			gmode = KGIGT_8BIT; /* 8 bit mode */
			break;
			case GR05_WRITEMODE1:
			gmode = KGIGT_1BIT; /* 1 bit mode */
			break;
			case GR05_WRITEMODE3:
			gmode = KGIGT_4BIT; /* 4 bit mode */
			break;
			case GR05_WRITEMODE2:
			break;
		}
		magy = (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	} else { /* Text mode */
		gmode = KGIGT_TEXT16;
		DCLKdiv = (reset_state.SEQ[0x01] & SR01_8DOT_CHCLK) ? 8 : 9;
	}
    
	/* Get horizontal timing */
	tx.width	= (reset_state.CRT[0x01]+1) * DCLKdiv;
	tx.blankstart	= reset_state.CRT[0x02] * DCLKdiv;
	tx.syncstart	= reset_state.CRT[0x04] * DCLKdiv;
	tx.syncend	= ((reset_state.CRT[0x05]) & CR05_HSE_MASK) * DCLKdiv;
	while (tx.syncend < tx.syncstart)
		tx.syncend += 0x20 * DCLKdiv;
	tx.blankend	= ((reset_state.CRT[0x03] & CR03_HBE_MASK) |
			   ((reset_state.CRT[0x05] & CR05_HBLANKEND_B5) ? 0x20 : 0) |
			   (reset_state.CRT[0x01] & 0xC0)) * DCLKdiv;
	tx.total	= (reset_state.CRT[0x00]+5) * DCLKdiv;
	tx.polarity	= (reset_state.MISC & MISC_NEG_HSYNC) != 0;

	/* Get vertical timing */
	ty.width	= 1 + (reset_state.CRT[0x12] | 
			       ((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B8) ? 0x100 : 0) |
			       ((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B9) ? 0x200 : 0));
	ty.blankstart	= 1 + (reset_state.CRT[0x15] |
			       ((reset_state.CRT[0x07] & CR07_VBLANKSTART_B8) ? 0x100 : 0) |
			       ((reset_state.CRT[0x09] & CR09_VBLANKSTART_B9) ? 0x200 : 0));
	ty.syncstart	= (reset_state.CRT[0x10] |                
			   ((reset_state.CRT[0x07] & CR07_VSYNCSTART_B8) ? 0x100 : 0) |
			   ((reset_state.CRT[0x07] & CR07_VSYNCSTART_B9) ? 0x200 : 0));
	ty.syncend	= 
	  ((reset_state.CRT[CRT_VDISPLAYEND] & ~CR11_VSYNCEND_MASK) |
	   (reset_state.CRT[CRT_VSYNCEND] & CR11_VSYNCEND_MASK)); 
	while (ty.syncend < ty.syncstart)
	  ty.syncend += (CR11_VSYNCEND_MASK + 1);

	ty.blankend	= 
	  (reset_state.CRT[0x16] | ((ty.width - 1) & ~0xFF)) + 1;
	ty.total = 2 + 
	  (reset_state.CRT[0x06] |
	   ((reset_state.CRT[0x07] & CR07_VTOTAL_B8) ? 0x100 : 0) |
	   ((reset_state.CRT[0x07] & CR07_VTOTAL_B9) ? 0x200 : 0));
	ty.polarity = (reset_state.MISC & MISC_NEG_VSYNC) != 0;
	
	dotclock = (CLK_in8(dpy, 0) == MISC_25MHZ_CLK) ? 25175000 : 28322000;
   
	/* Set all the mode info that we've just gathered */
	default_kgi_modereq.dpp.x	   = DCLKdiv;
	default_kgi_modereq.dpp.y	   = (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	default_kgi_modereq.frames	   = 1;
	default_kgi_modereq.visible.x = tx.width / default_kgi_modereq.dpp.x;
	default_kgi_modereq.visible.y = ty.width / default_kgi_modereq.dpp.y;
	default_kgi_modereq.virt.x	   = tx.width / default_kgi_modereq.dpp.x;
	default_kgi_modereq.virt.y	   = ty.width / default_kgi_modereq.dpp.y;
	default_kgi_modereq.size.x	   = 0;
	default_kgi_modereq.size.y	   = 0;
	default_kgi_modereq.graphtype = gmode;
	if (dpy->defmode == NULL)
		dpy->defmode = &default_kgi_modereq;
	else
		(*dpy->defmode) = default_kgi_modereq;

	dpy->mode.request	= default_kgi_modereq;
	dpy->mode.x		= tx;
	dpy->mode.y		= ty;
	dpy->mode.magnify.x	= magx;
	dpy->mode.magnify.y	= magy;
	dpy->mode.dclk		= dotclock;
	dpy->mode.clk.mul	= 1;
	dpy->mode.clk.div	= 1;
	dpy->mode.lclk.mul	= 1;
	dpy->mode.lclk.div	= 1;
	dpy->mode.chipset	= 0;
	
	default_kgi_mode = dpy->mode;

	chipset_setup_mode(dpy, &default_kgi_mode);
	chipset_setup_dpy(dpy, &(default_kgi_mode));

	return EOK;
}
   
static int chipset_init(struct kgi_display *dpy)
{
	int err;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	ASSERT(dpy);

	/* The following will provide it's own debug message */
	if (kgi_vga_dpy_init(dpy)) return(-E(CHIPSET,FAILED));

#define VGAMEM (((struct kgi_vgapriv *)(dpy->card->chipset->priv))->curr_mmio)

	if (detect_VGA(dpy->card->chipset) == EOK) {
	        NOTICE("Detected %dKBytes of Video RAM",
		       (dpy->card->chipset->ramsize/1024));
		if ((err = io_check_region(&vga_io))) {
		        DEBUG1("Final PIO region check failed (race?)");
		        outbMISC(reset_state.MISC); /* Altered by probe */
			return err;
		}

VGAMEM.decode = -1;
VGAMEM.baseptr = MEM_NULL;
VGAMEM.base = fb_text = fb_graph = 0x000A0000;
VGAMEM.size = fb_text_len = fb_graph_len = 0x00010000;
VGAMEM.name = "VGA MMIO";

		if ((err = mem_check_region(&VGAMEM))) {
		        outbMISC(reset_state.MISC); /* Altered by probe */
			return err;
		}

		io_claim_region(&vga_io);
		mem_claim_region(&VGAMEM);

#undef VGAMEM

		text16_data_init();
		return chipset_get_reset_mode(dpy);
	}

	return -E(CHIPSET, FAILED);
}

static void chipset_reset(struct kgi_display *dpy)
{
	restore_state(&reset_state);
	if (dpy) {
		dpy->mode = default_kgi_mode;
		(*dpy->defmode) = default_kgi_modereq;
	}
}

static void chipset_done(struct kgi_display *dpy)
{
	ASSERT(dpy);
	ASSERT(dpy->card);
	ASSERT(dpy->card->chipset);
	ASSERT(dpy->card->chipset->priv);

	chipset_reset(dpy);

	io_free_region(&vga_io);
	mem_free_region
	  (&(((struct kgi_vgapriv *)(dpy->card->chipset->priv))->curr_mmio));
	kgi_vga_dpy_done(dpy);
}

static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == KGIGT_TEXT16)
		? mode->request.dpp.x : 8;

#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo)
		mode->x.total += foo - (mode->x.total % foo);

	if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
		mode->x.syncend -= foo;
	if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
		mode->x.blankend -= foo;

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
		mode->y.syncend--;
	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
		mode->y.blankend--;

#undef ADJUST
}

/*
** rounds up virt and visible fields to the next multiple of 'round'.
** Enlarges virt to be at least of size visible.
*/
static void roundmode(kgi_modereq *gm,int roundx,int roundy)
{
	gm->visible.x = ((gm->visible.x + roundx - 1) / roundx) * roundx;
	gm->visible.y = ((gm->visible.y + roundy - 1) / roundy) * roundy;
	gm->virt.x    = ((gm->virt.x + roundx - 1) / roundx) * roundx;
	gm->virt.y    = ((gm->virt.y + roundy - 1) / roundy) * roundy;

	if (gm->visible.x > gm->virt.x)
 	        gm->virt.x = gm->visible.x;
	if (gm->visible.y > gm->virt.y)
	        gm->virt.y = gm->visible.y;
}

static int chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
                              enum kgi_tm_cmd cmd)
{
	int bpp = 0;
	int textx, texty;
	textx = mode->request.dpp.x;
	texty = mode->request.dpp.y;

	ASSERT(dpy);
	ASSERT(dpy->card);
	ASSERT(dpy->card->chipset);
	ASSERT(dpy->card->chipset->priv);

	switch (mode->request.graphtype) 
	{
	case KGIGT_8BIT:
		bpp = 8;	/* 1 Bpp */
		break; 	       
	case KGIGT_1BIT:
	case KGIGT_4BIT:
		bpp = 4;	/* 0.5 Bpp */
		break;	
	case KGIGT_TEXT16:
		break;		/* Not used in textmode */
	default:
		DEBUG0("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET,FAILED);
	}

#define XVIRTUAL (mode->request.virt.x)
#define YVIRTUAL (mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

	switch (cmd) {
	case CMD_PROPOSE:
		if (mode->request.graphtype == KGIGT_TEXT16) 
		{

			mode->magnify.x = mode->magnify.y = 1;
			if ((textx < 8) || (9 < textx)) 
			{
				DEBUG0("Invalid char width (%i)", textx);
				return -EINVAL;
			}

			if (XVIRTUAL == 0 && YVIRTUAL == 0)
			  return -EINVAL;
			
			if (XVIRTUAL == 0) 
			  XVIRTUAL = fb_text_len / (YVIRTUAL * 2);
			
			if (YVIRTUAL == 0)
			  YVIRTUAL = fb_text_len / (XVIRTUAL * 2);

			roundmode(&mode->request, 1, 1);
			if ((2 * XVIRTUAL * YVIRTUAL) > fb_text_len) {
				NOTICE("Not enough text memory");
				return -ENOMEM;
			}
			mode->x.width = mode->request.visible.x * textx;
			mode->y.width = mode->request.visible.y * texty;

		} else {	/* mode->request.graphtype == KGIGT_TEXT16 */

			if (XVIRTUAL == 0 && YVIRTUAL == 0)
			  return -EINVAL;

			if (XVIRTUAL == 0)
			  XVIRTUAL = fb_graph_len / YVIRTUAL;

			if (YVIRTUAL == 0)
			  YVIRTUAL = fb_graph_len / XVIRTUAL;

			mode->magnify.x = 1;
			mode->magnify.y = 1;
			roundmode(&mode->request, 8, 1);

			while ((XVISIBLE * mode->magnify.x) < 640) 
			  mode->magnify.x++;

			while ((YVISIBLE * mode->magnify.y) < 350) 
			  mode->magnify.y++;
			
			/* Can't zoom more than this */
			if (mode->magnify.y > 2)
			  return -EINVAL;
			
			/* Can't do more than doublescan */
			if (mode->magnify.x > 2)
			  return -EINVAL;

			if ((XVIRTUAL * YVIRTUAL * bpp / 8) > 
			    dpy->card->chipset->ramsize)
			{
				if ((XVISIBLE * YVISIBLE * bpp / 8) > 
				    dpy->card->chipset->ramsize) 
				{
					DEBUG0("Not enough memory to do virtual %ix%i at %i bpp",
					       XVIRTUAL, YVIRTUAL, bpp);
					return -ENOMEM;
				} 
				XVIRTUAL = XVISIBLE;
				YVIRTUAL = YVISIBLE;
			}
			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
			mode->x.width = 
			  mode->request.visible.x * mode->magnify.x;
			mode->y.width = 
			  mode->request.visible.y * mode->magnify.y;
		}

		return EOK;
		
	case CMD_LOWER:

		if (mode->dclk < chipinfo.dclk.min) 
		{
			
			DEBUG1("DCLK too low (%i)", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
		mode->dclk = (mode->dclk > chipinfo.dclk.max)
		  ? chipinfo.dclk.max : mode->dclk;

		adjust_timing(mode);
		return EOK;

	case CMD_RAISE:

		if (mode->dclk > chipinfo.dclk.max) 
		{
			
			DEBUG1("DCLK too high (%i)", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
		mode->dclk = (mode->dclk < chipinfo.dclk.min)
		  ? chipinfo.dclk.min : mode->dclk;

		adjust_timing(mode);
		return EOK;

	case CMD_CHECK:
		if (mode->lclk.mul != 1) 
		{
			
			DEBUG1("Invalid lclk multiplier %d", mode->lclk.mul);
			return -E(CHIPSET,FAILED);
		}
		switch (mode->lclk.div) 
		{
		case 1:
			break;
		default:
			DEBUG1("Invalid CLK to LCLK divider %d", 
			       mode->lclk.div);
			return -E(CHIPSET,FAILED);
		}
		DEBUG2("xv %d yv %d\n",XVIRTUAL,YVIRTUAL);
		if (mode->request.graphtype == KGIGT_TEXT16) 
		{
			mode->flags |= FL_CHIP_COLORTEXT;
			if ((mode->magnify.x != 1) || (mode->magnify.y != 1)
			    || (textx < 8) || (9 < textx) || (texty < 1)
			    || (texty > 32) || 
			    ((2 * XVIRTUAL * YVIRTUAL) > fb_text_len)
			    || (mode->dclk > chipinfo.dclk.max)
			    || (mode->dclk < chipinfo.dclk.min))
			{
				NOTICE1("HW check failed");
				return -E(CHIPSET,FAILED);
			}
		} else {
			switch (XVIRTUAL) {
			case 320:
			case 360:
			case 640:
			case 720:
				break;
			default:
				return -E(CHIPSET,FAILED);
			}

			if (((XVIRTUAL * YVIRTUAL * bpp / 8) > 
			     dpy->card->chipset->ramsize)
			    || (mode->magnify.x > 2) || (mode->magnify.y > 2)
			    || (mode->clk.mul != 1) || (mode->clk.div != 1)
			    || (mode->dclk < chipinfo.dclk.min)
			    || (chipinfo.dclk.max < mode->dclk)
			    || (mode->magnify.x == 2 && 
				mode->request.graphtype != KGIGT_8BIT))
			  return -E(CHIPSET,FAILED);
		}

#define TM(X) ((chipinfo.maxdots.x < mode->x.X) || (chipinfo.maxdots.y < mode->y.X))
		if (TM(width) || TM(blankstart) || TM(syncstart)
		    || TM(syncend) || TM(blankend) || TM(total)) 
		{
			NOTICE1("Timing check failed");
			DEBUG1("Maxdots are (%d,%d)", chipinfo.maxdots.x,
			       chipinfo.maxdots.y);
			return -E(CHIPSET,FAILED);
		}
#undef TM

		chipset_setup_mode(dpy, mode);
		return EOK;

	default:
		ERROR("Invalid command in chipset_check_mode()!");
		INTERNAL_ERROR;
		return -E(CHIPSET,FAILED);
	}
	
#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

static void chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	struct statetable setup = default_state;
	register int foo;
	int DCLKdiv;
	int textx,texty;
	textx = mode->request.dpp.x;
	texty = mode->request.dpp.y;

	ASSERT(dpy);
	ASSERT(dpy->card);
	ASSERT(dpy->card->chipset);
	ASSERT(dpy->card->chipset->priv);

	/* Set up bits per pixel and logical screen width */

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
		foo = mode->request.virt.x * 2 / 4;
		       /* Due to the addressing mode used */
		break;
		case KGIGT_1BIT:
		case KGIGT_4BIT:
		foo = mode->request.virt.x >> 4;
		break;
		case KGIGT_8BIT:
		foo = mode->request.virt.x >> 3;
		break;
		default:
		ERROR("Unsupported mode in chipset_set_mode()!");
		INTERNAL_ERROR;
		return;
	}

	/* Adjust BITS macros to use "setup" as the stored state. */
	#define WITH_STATE setup

	setup.CRT[CRT_LOGICALWIDTH] = foo & 0xFF;

	if (mode->request.graphtype == KGIGT_TEXT16) {
		setup.CRT[CRT_MODE] = 
		  CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC | 
		  CR17_CGA_BANKING | CR17_HGC_BANKING;

		setup.CRT[CRT_UNDERLINE] |= CR14_UNDERLINE_MASK;

		SEQBITS((SR02_PLANE0 | SR02_PLANE1),
			     SR02_PLANE_MASK, SEQ_WPLANE);

		setup.GRC[GRC_MODECONTROL] |= GR05_ODD_EVEN;
		setup.GRC[GRC_MMAP_MODE] |= GR06_CHAIN_OE;

		GRCBITS(GR06_MAP_A0_64, GR06_MAP_MASK,
			     GRC_MMAP_MODE);

		setup.ATC[ATC_MODECONTROL] &= ~ 
		  (AR10_GRAPHICS | AR10_MONOMODE | 
		   0x10 /* reserved (0) */ | AR10_8BIT_MODE | AR10_ENB_BLINK);
		setup.ATC[ATC_MODECONTROL] |= (AR10_TOP_PAN | AR10_SELECT_B54);
		setup.ATC[ATC_BORDER] = 0;
		setup.ATC[ATC_HPAN] = 0;
		setup.ATC[ATC_PIXELPAD] &= ~ AR14_DATA_MASK;

		setup.SEQ[SEQ_MEMMODE] |= SR04_256K_ACCESS;

		switch ((DCLKdiv = textx)) {
			case 8:
			setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;
			setup.ATC[ATC_MODECONTROL] &= ~AR10_9DOT_CHARS;
			break;
			case 9:
 			setup.SEQ[SEQ_CLOCK] &= ~SR01_8DOT_CHCLK;
			ATCBITS(AR10_9DOT_CHARS,
				(AR10_9DOT_CHARS | AR10_TOP_PAN),
				ATC_MODECONTROL);
	 		break;
		 	default:
			ERROR("Unsupported textx in chipset_set_mode()!");
			INTERNAL_ERROR;
			return;
		}
		CRTBITS((texty - 1), CR09_CHARHEIGHT_MASK, CRT_CHARHEIGHT);

		ATCBITS(((textx == 9) ? 0x08 : 0x00), 
			AR13_DATA_MASK, ATC_HPAN); 
		
	} else {	/* mode->request.graphtype == KGIGT_TEXT16 */
		DCLKdiv = 8;

		setup.CRT[CRT_MODE] = 
		  CR17_CGA_BANKING | CR17_HGC_BANKING
		  | CR17_ENABLE_SYNC | CR17_CGA_ADDR_WRAP;

		for (foo = 0; foo < 16; foo++)
			setup.ATC[foo] = default_state.ATC[foo];

		setup.SEQ[SEQ_WPLANE] |= SR02_PLANE_MASK;

		SEQBITS(SR04_256K_ACCESS, 0x0f, SEQ_MEMMODE);

		GRCBITS((GR06_GRAPHMODE | GR06_MAP_A0_64),
			(GR06_GRAPHMODE | GR06_MAP_MASK),
			GRC_MMAP_MODE);

		setup.CRT[CRT_HBLANKEND] |= CR03_IS_VGA;

		CRTBITS((mode->magnify.y - 1), CR09_CHARHEIGHT_MASK,
			CRT_CHARHEIGHT);

		switch (mode->request.graphtype) 
		{
		case KGIGT_1BIT:
		case KGIGT_4BIT:
			if (mode->request.graphtype == KGIGT_1BIT) 
			{
				setup.SEQ[SEQ_WPLANE] |= SR02_PLANE_MASK
				  /* Enable plane 0 for CPU */;
				GRCBITS(0x0e, GR01_DATA_MASK, GRC_ENABLE);
				GRCBITS(0x00, GR04_PLANEMASK, GRC_READPLANE);
				GRCBITS(GR05_WRITEMODE0, 0x03, 
					GRC_MODECONTROL); 
			} else 
			{
			  /* Enable all planes for set/reset (color banking) */
			  GRCBITS(0x0f, GR01_DATA_MASK, GRC_ENABLE);
			  GRCBITS(GR05_WRITEMODE3, 0x03, GRC_MODECONTROL); 
			}

			setup.SEQ[SEQ_MEMMODE] |= 
			  SR04_NO_ODDEVEN | SR04_256K_ACCESS;
			setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;
			setup.CRT[CRT_MODE] |= CR17_BYTE_MODE;
			setup.ATC[ATC_MODECONTROL] |= AR10_GRAPHICS;

			break;

		case KGIGT_8BIT:

			/* Note that this implies multiply.x == 2 */
			setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;

			setup.SEQ[SEQ_WPLANE] |= SR02_PLANE_MASK;

			if (mode->request.virt.x == 320 &&
			    mode->request.virt.y == 200) 
			{
				/* Planar */
				SEQBITS((SR04_NO_ODDEVEN | SR04_CHAINED),
					0x0f, SEQ_MEMMODE); 
				setup.CRT[CRT_UNDERLINE] |= 
				  CR14_DOUBLEWORDMODE;
			} else 
			{ 
			  /* Mode X */
				SEQBITS(SR04_NO_ODDEVEN, 0x0f, SEQ_MEMMODE);
				setup.CRT[CRT_MODE] |= CR17_BYTE_MODE;
			}

			setup.ATC[ATC_MODECONTROL] &= ~ 
			  (AR10_GRAPHICS | AR10_MONOMODE | AR10_ENB_BLINK | 
			   0x10 /* reserved (0) */ | AR10_SELECT_B54);
			setup.ATC[ATC_MODECONTROL] |= 
			  (AR10_GRAPHICS | AR10_8BIT_MODE | AR10_TOP_PAN);
			setup.ATC[ATC_BORDER] = 0;
			setup.ATC[ATC_HPAN] = 0;
			setup.ATC[ATC_PIXELPAD] &= ~ AR14_DATA_MASK;

			setup.GRC[GRC_MODECONTROL] |= GR05_8BIT_COLOR;
			setup.GRC[GRC_BITMASK] = 0xff;	/* Enable all bits */

			break;
		default:
			INTERNAL_ERROR;
			return;
		}
	}
	
	/* Set up horizontal timing */
	
	foo = (mode->x.width / DCLKdiv) - 1;	/* Horizontal display end */
	setup.CRT[CRT_HDISPLAYEND] = foo & 0xFF;
	
	foo = (mode->x.blankstart / DCLKdiv);	/* Horizontal blank start */
	setup.CRT[CRT_HBLANKSTART] = foo & 0xFF;

	foo = (mode->x.syncstart / DCLKdiv);	/* Horizontal sync start  */
	setup.CRT[CRT_HSYNCSTART] = foo & 0xFF;

	foo = (mode->x.syncend / DCLKdiv);	/* Horizontal sync end    */
	CRTBITS(foo, CR05_HSE_MASK, CRT_HSYNCEND);

	foo = (mode->x.blankend / DCLKdiv);	/* Horizontal blank end   */
	CRTBITS(foo, CR03_HBE_MASK, CRT_HBLANKEND); 
	CRTBITS(((foo & 0x20) ? CR05_HBLANKEND_B5 : 0),
		CR05_HBLANKEND_B5, CRT_HSYNCEND); 

	foo = (mode->x.total / DCLKdiv) - 5;	/* Horizontal total       */
	setup.CRT[CRT_HTOTAL] = foo & 0xFF;

	/* Set up vertical timing */

	foo = mode->y.width - 1;		/* Vertical display end   */
	setup.CRT[CRT_VDISPLAYEND] = foo & 0xFF;
	CRTBITS((((foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0) |
		 ((foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0)),
		(CR07_VDISPLAYEND_B8 | CR07_VDISPLAYEND_B9), 
		CRT_OVERFLOW); 
	
	foo = mode->y.blankstart - 1;		/* Vertical blank start   */
	setup.CRT[0x15] |= foo & 0xFF;
	CRTBITS(((foo & 0x100) ? CR07_VBLANKSTART_B8 : 0),		    
		CR07_VBLANKSTART_B8, CRT_OVERFLOW); 
        CRTBITS(((foo & 0x200) ? CR09_VBLANKSTART_B9 : 0),
		CR09_VBLANKSTART_B9, CRT_CHARHEIGHT); 

	foo = mode->y.syncstart;		/* Vertical sync start    */
	setup.CRT[CRT_VSYNCSTART] = foo & 0xFF;
	CRTBITS((((foo & 0x100) ? CR07_VSYNCSTART_B8 : 0) |
		 ((foo & 0x200) ? CR07_VSYNCSTART_B9 : 0)),
		(CR07_VSYNCSTART_B8 | CR07_VSYNCSTART_B9), 
		CRT_OVERFLOW); 

	foo = mode->y.syncend;		        /* Vertical sync end      */
	CRTBITS(foo, CR11_VSYNCEND_MASK, 0x11);

	foo = mode->y.blankend; 		/* Vertical blank end     */
	setup.CRT[CRT_VBLANKEND] = foo & 0xFF;

	foo = mode->y.total - 2;		/* Vertical total         */
	setup.CRT[CRT_VTOTAL] = foo & 0xFF;
	CRTBITS((((foo & 0x100) ? CR07_VTOTAL_B8 : 0) |
		 ((foo & 0x200) ? CR07_VTOTAL_B9 : 0)),
		(CR07_VTOTAL_B8 | CR07_VTOTAL_B9), 
		CRT_OVERFLOW); 

	MISCBITS((((mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC) |
		  ((mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC)),
		 (MISC_NEG_HSYNC | MISC_NEG_VSYNC));

	/* Set splitline to 0 */

	setup.CRT[CRT_LINECOMPARE] = 0;
	CRTBITS(0, CR07_LINECOMPARE_B8, CRT_OVERFLOW);
	CRTBITS(0, CR09_LINECOMPARE_B9, CRT_CHARHEIGHT);

	/* Ok.  We'll do the actual hardware programming now. */
	
	kgim_monitor_set_mode(dpy, mode);
	kgim_ramdac_set_mode(dpy, mode);
	kgim_clock_set_mode(dpy, mode);

	restore_state(&setup);

	chipset_setup_dpy(dpy, mode);

	NOTICE1("Mode set complete");
}

static int chipset_command(struct kgi_graphic *state,kgi_uint cmd,kgi_uint arg)
{
	kgi_coord  *coord;
	struct kgi_font *font;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
		case CHIP_SETDEFFONT:
		font = (struct kgi_font *)(state->io_buf);
		text16_set_font(dpy, font, 0, font->info->positions);
		return 0;
		case CHIP_SETSPLITLINE:
			if (arg > state->dev.mode.visible.y)
				return -E(CHIPSET,BOGUS);

			if (state->dev.org[0].y > state->dev.mode.visible.y - arg )
				kgi_set_origin(&(state->dev), 0, 
						state->dev.org[0].x,
						state->dev.mode.visible.y - arg );

			kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (kgi_coord *)(state->io_buf);

			if (coord->x < 0 || 
			    coord->x > state->dev.mode.virt.x - state->dev.mode.visible.x)
				return -E(CHIPSET,BOGUS);

			if (coord->y < 0 || dpy->mode.magnify.y<1 ||
			    coord->y > state->dev.mode.virt.y - state->dev.split/dpy->mode.magnify.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}

	return -E(CHIPSET,UNKNOWN);
}

/* KGI Intraface wrappers */

kgi_sint kgim_chipset_init(struct kgi_card *card,struct kgi_display *dpy)
{
	int err;

	ASSERT(card);
	ASSERT(dpy);

	if ((err = chipset_init(dpy)) != EOK) return err;
	ASSERT (card->chipset);

	strcpy(card->manufact, card->chipset->manufact);
	strcpy(card->model, card->chipset->model);
	card->modes = card->chipset->modes;
	card->flags = card->chipset->flags;
	card->maxdots = card->chipset->maxdots;

/*	card->preset = default_kgi_mode; */

	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	chipset_done(dpy);
}

kgi_sint  kgim_chipset_check_mode
(struct kgi_display *dpy, struct kgi_mode *mode,
 enum kgi_tm_cmd *cmd)
{
	DEBUG1("Checking mode... (%dx%d %d)\n", mode->request.virt.x,
	       mode->request.virt.y, mode->request.graphtype);
	return chipset_check_mode(dpy, mode, *cmd);
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
  DEBUG1("Setting mode... (%dx%d %d)\n", mode->request.virt.x,
	 mode->request.virt.y, mode->request.graphtype);
	chipset_set_mode(dpy,mode);
}

kgi_sint kgim_chipset_command(struct kgi_graphic *state, kgi_uint cmd, kgi_uint iobuf)
{
	return chipset_command(state, cmd, iobuf);
}
